(*+
 *	File:		UVUAssist.inc2.p
 *
 *	Contains:	VU assistance methods
 *
 *	Written by:	David Shayer
 *
 *	Copyright:	 1988-90 by Apple Computer, Inc., all rights reserved.
 *
 *  Version: 1.0d6
 *
 *	Change History:
 *
 *		  1/10/90	DAS		  new today
 *
 *	To Do:
-*)

{Historical Note:  References to the "Mole" are equivalent to references to "Agent VU".
"Agent VU" was previously called the "Mole".}


{ The mole sees MacApp windows as plain windows, that are empty.  MacApp
doesn't put it's controls in the controllist, so VU won't even see standard
control manager controls.  The assistance proc updates the mole buffer before
it is sent back to the VU host, telling it that there are
items in the windows, and telling it about each item.  We tell it that the
windows are dialogs, because VU can do more with dialogs.  It knows about
static text and edit text in dialogs, in addition to controls. }

{ The way we handle TGridViews deserves a little explanation.  The mole hook
deals with 3 kinds of user interface gadgets.  Control Manager control, Dialog
Manager items, and Grid items.  

Control Manager controls include buttons, radio buttons, check boxes, scroll bars,
and popup menus.  What?, I hear you say.  Popup menus aren't controls.  The only
way VU can deal with popup menus is if they are written using the popup menu CDEF
originally supplied with the Communications Toolbox, which makes popup menus look
like controls.  Since VU expects popups to be controls, we make them controls.

Dialog Manager items includes controls, and also static and edit text items,
Icons, Pictures, and Grid items.  Edit text items include TEditText and TTEView, 
and their descendents.  We include TTEView as an edit text item because its the
closest thing to a TextEdit record that VU knows about.  

TGridViews are a problem because unlike all the other items mentioned above,
where there is exactly one Pascal Object for each user interface item, there
are many user interface items for a single TGridView object.  So there is special
case code for TGridViews.  Each cell in a grid is passed back as a single
user item. }

{========================================== TVUAssist ================================}

{$S VUAssistSeg}

{The MoleHook is declared as type C, which means it must be an external function.
In C, you can declare a C function to be type Pascal, and the compiler simply
uses C calling conventions for that function.  The Pascal compiler is too braindead
to do this.  Instead, when you declare a function to be type C, it must be an external
function.  You can't write the bloody function in Pascal.  The actual code for
MoleHook is in assembly.  All it does is accept the call from the mole
with a C interface, and then call TVUAssist.MoleAssist with a Pascal interface.
It seemed like a good idea at the time...}
FUNCTION MoleHook (Select :integer; Input :Ptr; VAR OutputSize :integer;
	Result :integer) :OSErr; C; EXTERNAL;


{ Init fMoleRefNum.  fMoleRefNum contains the reference number of the mole driver,
or 0 if no mole driver has been found yet.}
PROCEDURE TVUAssist.IVUAssist;
BEGIN
	{$IFC qDebug}
		gVUDebug:=false;
		DebugFlag(@gVUDebug,'9', NIL, AtStr('VU Assist'));
	{$ENDC}

	fMoleRefNum:=0;
	SELF.OpenMoleDriver;
END;


{ Try to open the mole driver if there is one. }
PROCEDURE TVUAssist.OpenMoleDriver;
VAR
	PBRec :ParamBlockRec;
	Err :integer;
	MoleName :Str15;
BEGIN
	MoleName:='.Mole';
	PBRec.ioCompletion:=NIL;
	PBRec.ioNamePtr:=@MoleName;
	PBRec.ioPermssn:=fsCurPerm;
	Err:=PBOpen (@PBRec, false);
	IF (PBRec.ioResult=noErr) THEN
	BEGIN
		{$IFC qDebug}
			WriteLn ('Mole Contacted');
		{$ENDC}
		fMoleRefNum:=PBRec.ioRefNum;
		SELF.ResumeMole;
	END;
END;


{ Call this routine when the application is suspended by Multifinder, to let
the mole know we are no longer the frontmost app. }
PROCEDURE TVUAssist.SuspendMole;
BEGIN
	SELF.SetDebuggerHook(NIL);
END;


{ Call this routine when MultiFinder puts us back into the foreground. }
PROCEDURE TVUAssist.ResumeMole; 
BEGIN
	SELF.SetDebuggerHook(@MoleHook);
END;


{ Tell the mole the address of our assistance procedure, 
so it can call us when it needs to. }
PROCEDURE TVUAssist.SetDebuggerHook (NewHook :ProcPtr); 
CONST
	SET_ASSIST = 100;
VAR
	PBRec :ParamBlockRec;
	Err :integer;
	LongPtr :^Longint;
BEGIN
	IF (fMoleRefNum<>0) THEN
	BEGIN
		PBRec.ioCompletion:=NIL;
		PBRec.ioVRefNum:=0;
		PBRec.ioRefNum:=fMoleRefNum;
		PBRec.csCode:=SET_ASSIST;
		LongPtr:=@PBRec.csParam[0];
		LongPtr^:=LongInt(NewHook);
		Err:=PBControl (@PBRec, false);
		IF (PBRec.ioResult<>noErr) THEN
			WriteLn ('Error sending Mole control call');
	END;
END;


{ This is the routine called by MoleHook.  It dispatchs out to the routines
which actually assist the mole. }
FUNCTION TVUAssist.MoleAssist (Select :integer; Input :Ptr; Output :Ptr; 
	VAR OutputSize :integer; IntResult :integer) :integer;
VAR
	Result		:MoleError;
	ReturnValue	:MoleError;
BEGIN
	Result:=MoleError(IntResult);
	ReturnValue:=Result;
	CASE MoleSelectors(Select) OF
		SMenuInfo:
			BEGIN
				ReturnValue:=SELF.DoMenuInfo (Input, Output, OutputSize, Result);
			END;
		SMenuItems:
			BEGIN
				ReturnValue:=SELF.DoMenuItems (Input, Output, OutputSize, Result);
			END;
		SetMItemKey:
			BEGIN
				{$IFC qDebug}
					IF gVUDebug THEN
						WriteLn ('* SetMItemKey');
				{$ENDC}
			END;
		SFrontWind:
			BEGIN
				{$IFC qDebug}
					IF gVUDebug THEN
						WriteLn ('* SFrontWind');
				{$ENDC}
			END;
		SFindWind:
			BEGIN
				ReturnValue:=SELF.DoFindWindowInfo (Input, Output, OutputSize, Result);
			END;
		SWindInfo:
			BEGIN
				ReturnValue:=SELF.DoSendWindowInfo (Input, Output, OutputSize, Result);
			END;
		SFindCtl:
			BEGIN
				ReturnValue:=SELF.DoFindControl (Input, Output, OutputSize, Result);
			END;
		SCtlInfo:
			BEGIN
				ReturnValue:=SELF.DoSendControlInfo (Input, Output, OutputSize, Result);
			END;
		SFindDItem:
			BEGIN
				{$IFC qDebug}
					IF gVUDebug THEN
						WriteLn ('* SFindDItem');
				{$ENDC}
			END;
		SDItemInfo:
			BEGIN
				ReturnValue:=SELF.DoSendDlogItemInfo (Input, Output, OutputSize, Result);
			END;
		OTHERWISE
			BEGIN
				{$IFC qDebug}
					IF gVUDebug THEN
						WriteLn ('* ????????   Select=', Select);
				{$ENDC} 
			END;
	END; {case} 
	MoleAssist:=integer(ReturnValue);
	{$IFC qDebug}
		IF gVUDebug THEN
		BEGIN
			Write ('Result=',integer(Result));
			WriteLn ('     MoleAssist=',integer(ReturnValue));
		END;
	{$ENDC}
END;


{------------------------------------------------------------------------------------}
{ the routines for handling individual mole requests.  Each routine handles
   exactly one mole request. }

{ Handle a request for specific menu items.  The Mole can get info on menus in the
menu bar, but we have to supply info on popup menus. }
FUNCTION TVUAssist.DoMenuItems (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
	Result :MoleError) :MoleError;
VAR
	MenuReqPtr		:SendMenuItemsParamsPtr;
	MenuDescPtr		:MoleDataBlockPtr;
	Popup			:TPopup;
BEGIN
	DoMenuItems:=Result;
	MenuReqPtr:=SendMenuItemsParamsPtr(Input);
	MenuDescPtr:=MoleDataBlockPtr(Output);
	
	{ we don't respond when VU finds menus by rank.  
	It only wants to know about popupswhen it finds by ID. }
	IF MenuReqPtr^.menu_ID<>0 THEN
	BEGIN { find menu by menu ID }
		IF (Result<>mNoErr) THEN
		BEGIN
			Popup:=SELF.MenuIDToPopup (MenuReqPtr^.menu_ID);
			IF Popup<>NIL THEN
			BEGIN
				DoMenuItems:=SELF.FillPopupItemDesc (Popup, RANK_OF_THE_INVISIBLE, 
					MenuReqPtr^.start_item, MenuReqPtr^.stop_item, MenuDescPtr, OutputSize);
			END;
		END;
	END;
	
	{$IFC qDebug}
		IF gVUDebug THEN
		BEGIN
			WriteLn ('SMenuItems');
			SELF.PrintMenuItemReq (MenuReqPtr);
			SELF.PrintMenuItemDesc (MenuDescPtr, OutputSize);
		END;
	{$ENDC}
END;


{ Handle a request for menu info.  The Mole can get info on menus in the
menu bar, but we have to supply info on popup menus. }
FUNCTION TVUAssist.DoMenuInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
	Result :MoleError) :MoleError;
VAR
	MenuReqPtr		:SendMenuInfoParamsPtr;
	MenuDescPtr		:MoleMenuInfoPtr;
	Popup			:TPopup;
BEGIN
	DoMenuInfo:=Result;
	MenuReqPtr:=SendMenuInfoParamsPtr(Input);
	MenuDescPtr:=MoleMenuInfoPtr(Output);
	
	{ we don't respond when VU finds menus by rank.  
	It only wants to know about popupswhen it finds by ID. }
	IF MenuReqPtr^.menu_ID<>0 THEN
	BEGIN { find menu by menu ID }
		IF (Result<>mNoErr) THEN
		BEGIN
			Popup:=SELF.MenuIDToPopup (MenuReqPtr^.menu_ID);
			IF Popup<>NIL THEN
			BEGIN
				SELF.FillPopupDesc (Popup, RANK_OF_THE_INVISIBLE, MenuDescPtr, OutputSize);
				DoMenuInfo:=mNoErr;
			END;
		END;
	END;
	
	{$IFC qDebug}
		IF gVUDebug THEN
		BEGIN
			WriteLn ('SMenuInfo');
			SELF.PrintMenuInfoReq (MenuReqPtr);
			SELF.PrintMenuInfoDesc (MenuDescPtr);
		END;
	{$ENDC}
END;


{ Don't do anything for SFindWind call.  When doing an SFindWind call, the mole
should call SWindInfo before it executes SFindWind, so the SWindInfo assistance proc 
should fill in the correct data. }
FUNCTION TVUAssist.DoFindWindowInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer;
	Result :MoleError) :MoleError;
VAR
	WindReqPtr	:SendFindWindowParamPtr;
	WindDescPtr	:MoleWDescPtr;
	Window		:TWindow;
BEGIN
	DoFindWindowInfo:=Result;
	WindReqPtr:=SendFindWindowParamPtr(Input);
	WindDescPtr:=MoleWDescPtr(Output);
	{$IFC qDebug}
		IF gVUDebug THEN
		BEGIN
			WriteLn ('SFindWind');
		{	SELF.PrintFindWindReq (WindReqPtr);
			SELF.PrintWindDesc (WindDescPtr);}
		END;
	{$ENDC}
END;


{ Fill in the fields of the MoleWDescPtr not filled in by the mole.
These are the controlCount and numItems fields.  Update validFlags and
windowKind. }
FUNCTION TVUAssist.DoSendWindowInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer;
	Result :MoleError) :MoleError;
VAR
	WindReqPtr	:DWindReqPtr;
	WindDescPtr	:MoleWDescPtr;
	Window		:TWindow;
BEGIN
	DoSendWindowInfo:=Result;
	WindReqPtr:=DWindReqPtr(Input);
	WindDescPtr:=MoleWDescPtr(Output);
	Window:=SELF.RankToWindow (WindReqPtr^);
	IF Window<>NIL THEN 
	BEGIN
		WindDescPtr^.validFlags := BOR (WindDescPtr^.validFlags, VCTLCNT_MASK);
		WindDescPtr^.validFlags := BOR (WindDescPtr^.validFlags, VITEMCNT_MASK);
		WindDescPtr^.windowKind:=dialogKind;
		WindDescPtr^.controlCount:=SELF.ViewCount (Window, false);
		WindDescPtr^.numItems:=SELF.ViewCount (Window, true);
		{ OutputSize is filled in by the mole }
		DoSendWindowInfo:=mNoErr;
	END;
	{$IFC qDebug}
		IF gVUDebug THEN
		BEGIN
			WriteLn ('SWindInfo');
			SELF.PrintWindReq (WindReqPtr);
			SELF.PrintWindDesc (WindDescPtr);
		END;
	{$ENDC}
END;


{ The mole sends us a point location in global coordinates, we need to figure out
which window its in, which control in that window, and send back info about
that control. }
FUNCTION TVUAssist.DoFindControl (Input :Ptr; Output :Ptr; VAR OutputSize :integer;
	Result :MoleError) :MoleError;
VAR
	CtrlReqPtr	:SendFindControlParamPtr;
	CtrlDescPtr	:MoleCDescPtr;
	TargetPoint	:SendFindControlParam;
	part		:integer;
	wind		:WindowPeek;
	SavePort	:GrafPtr;
	Control		:TControl;
	Rank		:integer;
	SaveCtlList	:ControlHandle;
	WhichCtl	:ControlHandle;
	Window		:TWindow;
	VPt			:VPoint;
BEGIN
	DoFindControl:=Result;
	CtrlReqPtr:=SendFindControlParamPtr(Input);
	TargetPoint:=CtrlReqPtr^;
	CtrlDescPtr:=MoleCDescPtr(Output);

	part:=FindWindow (TargetPoint, WindowPtr(wind));
	{ The mole also tests for part=inSysWindow, but we'll let the mole handle those cases}
	IF (wind<>NIL) AND (part=inContent) THEN
	BEGIN
		Window:=gApplication.WMgrToWindow(WindowPtr(wind));
		IF (Window<>NIL) THEN
		BEGIN
			Control:=SELF.PointToControl (Window, TargetPoint, Rank);
			IF Control<>NIL THEN
			BEGIN
				part:=-1;
				IF (Member (Control, TCtlMgr)) THEN
				BEGIN
					{ Get the control part code.  Controls are not kept in the controllist,
					so we temporarily put them in there.  Also, they are in their subviews
					local coordinate system, so we must convert targetpoint. }
					GetPort (SavePort);
					SetPort (WindowPtr(wind));
					IF Window.Focus THEN
					BEGIN
						SaveCtlList:=wind^.controlList;
						wind^.controlList:=TCtlMgr(Control).fCMgrControl;
						SELF.GlobalToWindow (Window, TargetPoint, VPt);
						SELF.WindowToView (Control, VPt, TargetPoint);
						part:=FindControl (TargetPoint, WindowPtr(wind), WhichCtl);
						wind^.controlList:=SaveCtlList;
					END;
					SetPort (SavePort);
				END;
	
				SELF.FillCtrlDesc (CtrlDescPtr, part, SELF.WindowToRank(wind), 
					Rank, Control, OutputSize);
				DoFindControl:=mNoErr;
			END;
		END;
	END;
	{$IFC qDebug}
		IF gVUDebug THEN
			BEGIN
				WriteLn ('SFindCtl');
				SELF.PrintCtrlPoint (CtrlReqPtr);
				SELF.PrintCtrlDesc (CtrlDescPtr); 
			END;
	{$ENDC}
END;


{ Read the input parameter to see which control they want info on, then
fill in the SendCtlInfoParamsPtr record.  The mole can't see any info about
controls, so we fill in the whole record, except for part, which doesn't apply.}
FUNCTION TVUAssist.DoSendControlInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer;
	Result :MoleError) :MoleError;
VAR
	CtrlReqPtr	:SendCtlInfoParamsPtr;
	CtrlDescPtr	:MoleCDescPtr;
	Window		:TWindow;
	Control		:TControl;
BEGIN
	DoSendControlInfo:=Result;
	CtrlReqPtr:=SendCtlInfoParamsPtr(Input);
	CtrlDescPtr:=MoleCDescPtr(Output);

	Window:=SELF.RankToWindow (CtrlReqPtr^.window_rank);
	IF Window<>NIL THEN
	BEGIN
		Control:=TControl(SELF.RankToItem (Window, CtrlReqPtr^.control_rank, false));
		IF Control<>NIL THEN
		BEGIN
			SELF.FillCtrlDesc (CtrlDescPtr, -1, CtrlReqPtr^.window_rank, 
				CtrlReqPtr^.control_rank, Control, OutputSize);
			DoSendControlInfo:=mNoErr;
		END;
	END;
	
	{$IFC qDebug}
		IF gVUDebug THEN
			BEGIN
				WriteLn ('SCtlInfo');
				SELF.PrintCtrlReq (CtrlReqPtr);
				SELF.PrintCtrlDesc (CtrlDescPtr);
			END;
	{$ENDC}
END;


{ Read the input to see which dialog item in which window VU wants data on, and
fill in the dialogietminfo record. }
FUNCTION TVUAssist.DoSendDlogItemInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer;
	Result :MoleError) :MoleError;
VAR
	DlogReqPtr	:SendDItemInfoPtr;
	DlogDescPtr	:MoleDItemInfoPtr;
	Window		:TWindow;
	View		:TView;
BEGIN
	DoSendDlogItemInfo:=Result;
	DlogReqPtr:=SendDItemInfoPtr(Input);
	DlogDescPtr:=MoleDItemInfoPtr(Output);
	
	Window:=SELF.RankToWindow (DlogReqPtr^.window_rank);
	IF Window<>NIL THEN
	BEGIN
		View:=SELF.RankToItem (Window, DlogReqPtr^.item_num, true);
		IF View<>NIL THEN
		BEGIN
			SELF.FillDlogItemDesc (DlogDescPtr, Window, DlogReqPtr^.window_rank, 
				DlogReqPtr^.item_num, View, OutputSize);
			DoSendDlogItemInfo:=mNoErr;
		END;
	END;

	{$IFC qDebug}
		IF gVUDebug THEN
			BEGIN
				WriteLn ('SDItemInfo');
				SELF.PrintDlogReq (DlogReqPtr);
				SELF.PrintDlogDesc (DlogDescPtr);
			END;
	{$ENDC}
END;


{------------------------------------------------------------------------------------}
{ routines for filling in mole data structures }

{ Given a TPopup object, fill in the mole's menu descriptor record. }
PROCEDURE TVUAssist.FillPopupDesc (Popup: TPopup; MenuRank: integer; 
	MenuDescPtr :MoleMenuInfoPtr; VAR OutputSize :integer);
BEGIN
	WITH MenuDescPtr^ DO
	BEGIN
		num_items:=CountMItems (Popup.fMenuHandle);
		rank:=MenuRank;
		left_edge:=0;
		menuID:=Popup.fMenuID;
		menuWidth:=Popup.fMenuHandle^^.menuWidth;
		menuHeight:=Popup.fMenuHandle^^.menuHeight;
		enableFlags:=Popup.fMenuHandle^^.enableFlags;
		text_name:=Popup.fMenuHandle^^.menuData;
		
		{ this ugly thing computes the length (bytes used) of MoleMenuInfo }
		OutputSize:=sizeof(MoleMenuInfo) - (255-length(text_name));
	END;
END;


{ Given a TPopup object, fill in the mole's menuitem descriptor record. }
FUNCTION TVUAssist.FillPopupItemDesc (Popup: TPopup; MenuRank: integer; start: integer; 
	stop: integer; MenuItemDescPtr :MoleDataBlockPtr; VAR OutputSize :integer) :MoleError;
	
	{ this function depends upon Ptr being a pointer to a single byte }
	FUNCTION SeekMenuItem (ItemNum: integer; MenuHdl: MenuHandle) :Ptr;
	VAR
		SeekPtr	:Ptr;
		SeekNum	:integer;
	BEGIN
		SeekNum:=1;
		SeekPtr:= Ptr(ORD4(@MenuHdl^^.menuData) + length(MenuHdl^^.menuData) + 1);
		WHILE (SeekNum<ItemNum) AND (SeekPtr^<>0) DO
		BEGIN
			SeekPtr:= Ptr(ORD4(SeekPtr) + SeekPtr^ + ADDITIONAL_MENU_ITEM_DATA);
			SeekNum:=SeekNum+1;
		END;
		SeekMenuItem:=SeekPtr;
	END;
	
VAR
	StartPtr	:Ptr;
	StopPtr		:Ptr;
BEGIN
	FillPopupItemDesc:=mNoErr;
	
	IF (start>CountMItems (Popup.fMenuHandle)) OR (start<1) THEN
	BEGIN
		FillPopupItemDesc:=mBadMenuItemIndex;
		EXIT (FillPopupItemDesc);
	END;
	
	IF stop>CountMItems (Popup.fMenuHandle) THEN
	BEGIN
		stop:=CountMItems (Popup.fMenuHandle);
		FillPopupItemDesc:=mBadLastMenuItemIndex;
	END;
	
	StartPtr:=SeekMenuItem (start, Popup.fMenuHandle);
	StopPtr:=SeekMenuItem (stop+1, Popup.fMenuHandle);
	IF (StartPtr=NIL) OR (StopPtr=NIL) THEN
	BEGIN
		FillPopupItemDesc:=mBadMenuItemIndex;
		EXIT (FillPopupItemDesc);
	END;
	MenuItemDescPtr^.blockLength:=stop-start+1;
	BlockMove (StartPtr, @MenuItemDescPtr^.dataBlock[0], ORD4(StopPtr)-ORD4(StartPtr));
	
	{ this ugly thing computes the length (bytes used) of MoleDataBlock }
	OutputSize:=sizeof(integer) + (ORD4(StopPtr)-ORD4(StartPtr));
END;


{ Fill in a mole control descriptor record }
PROCEDURE TVUAssist.FillCtrlDesc (CtrlDescPtr :MoleCDescPtr; thePart :integer; 
	WindRank :integer; ControlRank :integer; Control :TControl; VAR OutputSize :integer);
VAR
	dummy		:longint;
	ControlHdl	:ControlHandle;
	aVRect		:VRect;
BEGIN
	IF Member (Control, TCtlMgr) THEN
	BEGIN
		ControlHdl:=TCtlMgr(Control).fCMgrControl;
		WITH CtrlDescPtr^, ControlHdl^^ DO
		BEGIN
			validFlags:=VALL_C_MASK;
			IF thePart=-1 THEN
			BEGIN
				dummy:=validFlags;
				BClr (dummy, VCPART);
				validFlags:=dummy;
			END;
			owner_rank:=WindRank;
			ctlRect:=contrlRect;
			SELF.ViewRectToWindowRect (ctlRect, Control);
			ctlValue:=contrlValue;
			ctlMin:=contrlMin;
			ctlMax:=contrlMax;
			ctlHilite:=contrlHilite;
			IF thePart=-1 THEN
				part:=0
			ELSE
				part:=thePart;
			ctlRank:=ControlRank;
			ctlDefID:=BAND(BSR(longint(contrlDefProc), 24), $7);
			IF Member (Control, TScrollBar) THEN
				ctlDefID:=ctlDefID+16;
			popupMenuID:=0;
			text_name:=contrlTitle;
	
			{ this ugly thing computes the length (bytes used) of CtrlDescPtr }
			OutputSize:=sizeof(MoleControlDescriptor) - (255-length(text_name));
		END;
	END
	ELSE
	IF Member (Control, TPopup) THEN
	BEGIN
		WITH CtrlDescPtr^ DO
		BEGIN
			validFlags:=VALL_C_MASK;
			IF thePart=-1 THEN
			BEGIN
				dummy:=validFlags;
				BClr (dummy, VCPART);
				validFlags:=dummy;
			END;
			owner_rank:=WindRank;
			Control.GetExtent (aVRect);
			VRectToRect(aVRect, ctlRect);
			SELF.ViewRectToWindowRect (ctlRect, Control);

			ctlValue:=TPopup(Control).GetCurrentItem;
			ctlMin:=1;
			ctlMax:=CountMItems(TPopup(Control).fMenuHandle);
			ctlHilite:=0;
			IF thePart=-1 THEN
				part:=0
			ELSE
				part:=thePart;
			ctlRank:=ControlRank;
			ctlDefID:=POPUP_CDEF*16;
			popupMenuID:=TPopup(Control).fMenuID;
			text_name:=TPopup(Control).fMenuHandle^^.menuData;
	
			{ this ugly thing computes the length (bytes used) of CtrlDescPtr }
			OutputSize:=sizeof(MoleControlDescriptor) - (255-length(text_name));
		END;
	END;
END;


PROCEDURE TVUAssist.ViewRectToWindowRect (VAR aRect :Rect; aView :TView);
VAR
	VPt :VPoint;
BEGIN
	PtToVPt(aRect.topLeft, VPt);
	aView.LocalToWindow (VPt);
	aRect.topLeft:=VPtToPt (VPt);
	PtToVPt(aRect.botRight, VPt);
	aView.LocalToWindow (VPt);
	aRect.botRight:=VPtToPt (VPt);
END;


{ Fill in a mole dialog item descriptor record }
PROCEDURE TVUAssist.FillDlogItemDesc (DlogDescPtr :MoleDItemInfoPtr; Window :TWindow;
	WindRank :integer; ItemRank :integer; Item :TView; VAR OutputSize :integer);
VAR
	aVRect		:VRect;
	TextLen 	:longint;
	TextHdl 	:Handle;
	ItemKind	:ViewItems;
	aCell		:GridCell;
	DisposeObj	:Boolean;
BEGIN
	DisposeObj:=false;
	WITH DlogDescPtr^ DO
	BEGIN
		ItemKind:=SELF.GetItemType (Item, true, true);
		{ fill in box, itemType, item_enabled, has_text and text_name here }
		CASE ItemKind OF
			VINotMember:
				;
			VIButton:
			BEGIN
				TCtlMgr(Item).ControlArea (box);
				itemType:=btnCtrlItem;
				item_enabled:=TCtlMgr(Item).IsCMgrVisible AND Item.IsViewEnabled;
				has_text:=TRUE;
				TCtlMgr(Item).GetText (text_name);
			END;
			VIRadio:
			BEGIN
				TCtlMgr(Item).ControlArea (box);
				itemType:=radCtrlItem;
				item_enabled:=TCtlMgr(Item).IsCMgrVisible AND Item.IsViewEnabled;
				has_text:=TRUE;
				TCtlMgr(Item).GetText (text_name);
			END;
			VICheckBox:
			BEGIN
				TCtlMgr(Item).ControlArea (box);
				itemType:=chkCtrlItem;
				item_enabled:=TCtlMgr(Item).IsCMgrVisible AND Item.IsViewEnabled;
				has_text:=TRUE;
				TCtlMgr(Item).GetText (text_name);
			END;
			VIScrollBar:
			BEGIN
				TCtlMgr(Item).ControlArea (box);
				itemType:=resCtrlItem;
				item_enabled:=TCtlMgr(Item).IsCMgrVisible AND Item.IsViewEnabled;
				has_text:=TRUE;
				TCtlMgr(Item).GetText (text_name);
			END;
			VIStatText:
			BEGIN
				Item.GetExtent (aVRect);
				VRectToRect(aVRect, box);
				itemType:=statText;
				item_enabled:=FALSE;
				has_text:=TRUE;
				TStaticText(Item).GetText (text_name);
			END;
			VIEditText:
			BEGIN
				Item.GetExtent (aVRect);
				VRectToRect(aVRect, box);
				itemType:=editText;
				item_enabled:=Item.IsViewEnabled;
				has_text:=TRUE;
				TStaticText(Item).GetText (text_name);
			END;
			VIIcon:
			BEGIN
				Item.GetExtent (aVRect);
				VRectToRect(aVRect, box);
				itemType:=iconItem;
				item_enabled:=Item.IsViewEnabled;
				has_text:=FALSE;
				item_text_ptr:=NIL;
			END;
			VIPopup:
			BEGIN
				Item.GetExtent (aVRect);
				VRectToRect(aVRect, box);
				itemType:=resCtrlItem;
				item_enabled:=Item.IsViewEnabled;
				has_text:=FALSE;
				item_text_ptr:=NIL;
			END;
			VIPicture:
			BEGIN
				Item.GetExtent (aVRect);
				VRectToRect(aVRect, box);
				itemType:=picItem;
				item_enabled:=Item.IsViewEnabled;
				has_text:=FALSE;
				item_text_ptr:=NIL;
			END;
			VITextEdit:
			BEGIN
				Item.GetExtent (aVRect);
				VRectToRect(aVRect, box);
				itemType:=editText;
				item_enabled:=Item.IsViewEnabled;
				has_text:=TRUE;
				TextHdl:=TTEView(Item).ExtractText;
				TextLen:=GetHandleSize (TextHdl);
				IF TextLen>255 THEN TextLen:=255;
				BlockMove (TextHdl^, @text_name, TextLen);
			END;
			VIGridItem:
			BEGIN
				Item.GetExtent (aVRect);
				VRectToRect(aVRect, box);
				itemType:=userItem;
				item_enabled:=TGridItem(Item).fGridView.IsViewEnabled;
				has_text:=TRUE;
				aCell.v:=TGridItem(Item).fRow;
				aCell.h:=TGridItem(Item).fColumn;
				TTextGridView(TGridItem(Item).fGridView).GetText (aCell, text_name);
				DisposeObj:=true;
			END;
		END; { case }
		SELF.ViewRectToWindowRect (box, Item);
		{ fill in ctlRank, owner_rank, and item_num here }
		ctlRank:=SELF.ItemToRank (Window, Item, false);
		owner_rank:=WindRank;
		item_num:=ItemRank;
			
		{ this ugly thing computes the length (bytes used) of MoleDItemInfo }
		IF ItemKind<>VINotMember THEN
			OutputSize:=sizeof(MoleDItemInfo) - (255-length(text_name));
		
		IF DisposeObj THEN
			Item.Free;
	END; { with }
END;


{------------------------------------------------------------------------------------}
{ routines for searching through the view tree to find particular views }

{ Given the rank (z-order) of a window, return the window object.  Invisible
windows are not considered to be in the window list, for ranking purposes. }
FUNCTION TVUAssist.RankToWindow (Rank :integer) :TWindow;
VAR
	i: integer;
	Window :WindowPeek;
BEGIN
	RankToWindow:=NIL;
	i:=1;
	Window:=WindowPeek(FrontWindow);
	WHILE Window<>NIL DO
	BEGIN
		IF Window^.visible THEN
		BEGIN
			IF i=Rank THEN
			BEGIN
				RankToWindow:=gApplication.WMgrToWindow(WindowPtr(Window));
				LEAVE;
			END;
			i:=i+1;
		END;
		Window:=Window^.nextWindow;
	END;
END;


{Given a window ptr, return the window's rank (z-order).}
FUNCTION TVUAssist.WindowToRank (WindToFind :WindowPeek) :integer;
VAR
	i: integer;
	Window :WindowPeek;
BEGIN
	WindowToRank:=0;
	i:=1;
	Window:=WindowPeek(FrontWindow);
	WHILE Window<>NIL DO
	BEGIN
		IF Window^.visible THEN
		BEGIN
			IF WindToFind=Window THEN
			BEGIN
				WindowToRank:=i;
				LEAVE;
			END;
			i:=i+1;
		END;
		Window:=Window^.nextWindow;
	END;
END;


{ Convert a point in global coordinates to window coordinates. }
PROCEDURE TVUAssist.GlobalToWindow (Window :TWindow; Pt :Point; VAR VPt :VPoint);
VAR
	SavePort	:GrafPtr;
	TargetPoint	:Point;
BEGIN
	VPt.h:=0;
	VPt.v:=0;
	GetPort (SavePort);
	SetPort (Window.GetGrafPort);
	IF Window.Focus THEN
	BEGIN
		TargetPoint:=Pt;
		GlobalToLocal (TargetPoint);
		Window.QDToViewPt(TargetPoint, VPt);
	END;
	SetPort (SavePort);
END;


{ Convert a point in window coordinates to a subview's coordinates. Return both a
VPoint and a QDPoint. }
PROCEDURE TVUAssist.WindowToView (View :TView; VAR VPt :VPoint; VAR Pt :Point);
BEGIN
	Pt.h:=0;
	Pt.v:=0;
	View.WindowToLocal(VPt);
	IF View.Focus THEN
		Pt:=View.ViewToQDPt (VPt);
END;



{ In the front (active) window, given the menu ID of a popup item, return the  
popup menu object and its rank. }
FUNCTION TVUAssist.MenuIDToPopup (MenuID :integer) :TPopup;
VAR
	Item 		:TView;
	Window		:TWindow;
	
	PROCEDURE DoToSubView(theSubView: TView);
	BEGIN
		IF theSubView.fShown THEN
		BEGIN	
			IF SELF.GetItemType (theSubView, false, false)=VIPopup THEN
			BEGIN
				IF TPopup(theSubView).fMenuID=MenuID THEN
					Item:=theSubView;
			END;
			IF Member (theSubView, TView) AND (Item=NIL) THEN
				theSubView.EachSubView (DoToSubView);
		END;
	END;
	
BEGIN
	Item:=NIL;
	Window:=gApplication.GetActiveWindow;
	IF (Window<>NIL) & (Window.fShown) THEN
		Window.EachSubView (DoToSubView);
	MenuIDToPopup:=TPopup(Item);
END;


{ Given a window object and a point in global coordinates,
find the control which is under that point. }
FUNCTION TVUAssist.PointToControl (Window :TWindow; Pt :Point; VAR Rank :integer) :TControl;
VAR
	Control 	: TControl;
	VPt			: VPoint;
	TestPoint	: VPoint;
	DummyPoint	: Point;
	aRect		: Rect;
	
	PROCEDURE DoToSubView(theSubView: TView);
	VAR
		ItemType :ViewItems;
	BEGIN
		IF theSubView.fShown THEN
		BEGIN	
			ItemType:= SELF.GetItemType (theSubView, false, false);
			IF ItemType<>VINotMember THEN
			BEGIN
				TestPoint:=VPt;
				SELF.WindowToView (theSubView, TestPoint, DummyPoint);
				IF theSubView.ContainsMouse(TestPoint) THEN
						Control:=TControl(theSubView);
			END;
			IF Member (theSubView, TView) AND (Control=NIL) THEN
				theSubView.EachSubView (DoToSubView);
		END;
	END;
	
BEGIN
	Control:=NIL;
	SELF.GlobalToWindow (Window, Pt, VPt);
	IF Window.fShown THEN
		Window.EachSubView (DoToSubView);
	Rank:=SELF.ItemToRank (Window, Control, false);
	PointToControl:=Control;
END;


{ Walk the view hierarchy tree, counting the number of user interface items.
Only visible views and items are counted. }
FUNCTION TVUAssist.ViewCount (Window :TWindow; DlogViewItems :Boolean) :integer;
VAR
	ItemCount :integer;
	
	PROCEDURE DoToSubView(theSubView: TView);
	VAR
		ItemType: ViewItems;
	BEGIN
		IF theSubView.fShown THEN
		BEGIN
			ItemType:=SELF.GetItemType (theSubView, DlogViewItems, DlogViewItems);
			IF ItemType<>VINotMember THEN
			BEGIN
				IF ItemType=VIGridItem THEN
					ItemCount:=ItemCount+SELF.CountGridItems (TGridView(theSubView))
				ELSE
					ItemCount := ItemCount + 1;
			END;
			IF Member (theSubView, TView) THEN
				theSubView.EachSubView (DoToSubView);
		END;
	END;
	
BEGIN
	ItemCount:=0;
	IF Window.fShown THEN
		Window.EachSubView (DoToSubView);
	ViewCount:=ItemCount;
END;


{Given a control or dialog item (View), and the window its in (Window), return its rank. }
FUNCTION TVUAssist.ItemToRank (Window :TWindow; View :TView; DlogViewItems :Boolean) :integer;
VAR
	ItemCount	:integer;
	Rank 		:integer;
	
	PROCEDURE DoToSubView(theSubView: TView);
	VAR
		ItemType:	ViewItems;
	BEGIN
		IF theSubView.fShown THEN
		BEGIN	
			ItemType:=SELF.GetItemType (theSubView, DlogViewItems, DlogViewItems);
			IF ItemType<>VINotMember THEN
			BEGIN
				IF ItemType=VIGridItem THEN
				BEGIN
					IF (Member(View, TGridItem)) & (theSubView=TGridItem(View).fGridView) THEN
						Rank:=ItemCount+TGridItem(View).fRank
					ELSE
						ItemCount:=ItemCount+SELF.CountGridItems (TGridView(theSubView));
				END
				ELSE BEGIN
					ItemCount := ItemCount + 1;
					IF theSubView=View THEN
						Rank:=ItemCount;
				END;
			END;
			IF Member (theSubView, TView) AND (Rank=0) THEN
				theSubView.EachSubView (DoToSubView);
		END;
	END;
	
BEGIN
	Rank:=0;
	ItemCount:=0;
	IF Window.fShown THEN
		Window.EachSubView (DoToSubView);
	ItemToRank:=Rank;
END;


{ Given a window object and the rank (z-order) of an item, return the item object. }
FUNCTION TVUAssist.RankToItem (Window :TWindow; Rank :integer; DlogViewItems :Boolean) :TView;
VAR
	ItemCount	:integer;
	Item 		:TView;
	
	PROCEDURE DoToSubView(theSubView: TView);
	VAR
		ItemType:	ViewItems;
		GridCount:	integer;
	BEGIN
		IF theSubView.fShown THEN
		BEGIN	
			ItemType:=SELF.GetItemType (theSubView, DlogViewItems, DlogViewItems);
			IF ItemType<>VINotMember THEN
			BEGIN
				IF ItemType=VIGridItem THEN
				BEGIN
					GridCount:=SELF.CountGridItems (TGridView(theSubView));
					IF (Rank>ItemCount) AND (Rank<=ItemCount+GridCount) THEN
						Item:=SELF.MakeGridItem (TGridView(theSubView), Rank-ItemCount);
					ItemCount := ItemCount + GridCount;
				END
				ELSE BEGIN
					ItemCount := ItemCount + 1;
					IF ItemCount=Rank THEN
						Item:=theSubView;
				END;
			END;
			IF Member (theSubView, TView) AND (Item=NIL) THEN
				theSubView.EachSubView (DoToSubView);
		END;
	END;
	
BEGIN
	Item:=NIL;
	ItemCount:=0;
	IF Window.fShown THEN
		Window.EachSubView (DoToSubView);
	RankToItem:=Item;
END;


{ Report the number of items in a TGridView.  This is here so it can
be easily overridden if you only want rows or columns to show up as
seperate user interface items, instead of each cell. }
FUNCTION TVUAssist.CountGridItems (theGridView: TGridView): integer;
BEGIN
	CountGridItems:=theGridView.fNumOfRows*theGridView.fNumOfCols;
END;

{ Instantiate a TGridItem object, and fill in its fields..  This is here 
so it can be easily overridden if you only want rows or columns to show 
up as seperate user interface items, instead of each cell. }
FUNCTION TVUAssist.MakeGridItem (theGridView: TGridView; Rank: integer): TGridItem;
VAR
	GridItem:	TGridItem;
BEGIN
	New (GridItem);
	FailNil (GridItem);
	GridItem.fGridView:=theGridView;
	GridItem.fRank:=Rank;
	GridItem.fRow:=(Rank-1) DIV theGridView.fNumOfCols +1;
	GridItem.fColumn:=(Rank-1) MOD theGridView.fNumOfCols +1;
	{$IFC qDebug}
		IF gVUDebug THEN
			GridItem.PrintInstanceVars;
	{$ENDC}
	MakeGridItem:=GridItem;
END;

{ Given a view item and the control/dialog item flag, decide if this
view object is a control if the flag=ControlsOnly, or if its a dialog
item if the flag=DialogItems, or if its neither.  Of course, MacApp 
doesn't have a concept of dialog items, so we improvise.  Here's the criteria.
Any descendent of TCtlMgr is a control.  Any descendent of TTView or
TControl (except for TCluster) is a dialog item. }
FUNCTION TVUAssist.GetItemType (Item :TView; DlogViewItems :Boolean;
	GridViewItems: Boolean) :ViewItems;
BEGIN
	GetItemType:=VINotMember;
	IF Member (Item, TButton) THEN
		GetItemType:=VIButton
	ELSE IF Member (Item, TRadio) THEN
		GetItemType:=VIRadio
	ELSE IF Member (Item, TCheckBox) THEN
		GetItemType:=VICheckBox
	ELSE IF Member (Item, TScrollBar) THEN
		GetItemType:=VIScrollBar
	ELSE IF Member (Item, TPopup) THEN	{ we consider popups to be controls, because }
		GetItemType:=VIPopup;			{ VU expects popups to use the popup CDEF }
	IF DlogViewItems THEN
	BEGIN
		IF Member (Item, TEditText) THEN	{TEditText descends from TStaticText}
			GetItemType:=VIEditText
		ELSE IF Member (Item, TStaticText) THEN
			GetItemType:=VIStatText
		ELSE IF Member (Item, TIcon) THEN
			GetItemType:=VIIcon
		ELSE IF Member (Item, TPicture) THEN
			GetItemType:=VIPicture
		ELSE IF Member (Item, TTEView) AND NOT Member (Item, TDialogTEView) THEN
			GetItemType:=VITextEdit;
	END;
	IF GridViewItems THEN
	BEGIN
		IF Member (Item, TGridView) THEN
			GetItemType:=VIGridItem;
	END;
END;


{------------------------------------------------------------------------------------}
{ debugging routines which print the contents of various mole data structures }

{$IFC qDebug}
PROCEDURE TVUAssist.PrintFindWindReq (WindReqPtr :SendFindWindowParamPtr);
BEGIN
	WriteLn ('WindReqPtr^=',WindReqPtr^.h, WindReqPtr^.v);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintWindReq (WindReqPtr :DWindReqPtr);
BEGIN
	WriteLn ('WindReqPtr^=',WindReqPtr^);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintWindDesc (WindDescPtr :MoleWDescPtr);
BEGIN
	Write ('WindDescPtr^validFlags=');
	WriteHexInt (WindDescPtr^.validFlags);
	WriteLn;
	WriteLn ('WindDescPtr^windowKind=',WindDescPtr^.windowKind);
	WriteLn ('WindDescPtr^portRect=',WindDescPtr^.portRect.top,
		WindDescPtr^.portRect.left, WindDescPtr^.portRect.bottom,
		WindDescPtr^.portRect.right);
	WriteLn ('WindDescPtr^structRect=',WindDescPtr^.structRect.top,
		WindDescPtr^.structRect.left, WindDescPtr^.structRect.bottom,
		WindDescPtr^.structRect.right);
	WriteLn ('WindDescPtr^contentRect=',WindDescPtr^.contentRect.top,
		WindDescPtr^.contentRect.left, WindDescPtr^.contentRect.bottom,
		WindDescPtr^.contentRect.right);
	WriteLn ('WindDescPtr^windowPart=',WindDescPtr^.windowPart);
	WriteLn ('WindDescPtr^windowRank=',WindDescPtr^.windowRank);
	WriteLn ('WindDescPtr^windowFlags=',WindDescPtr^.windowFlags);
	WriteLn ('WindDescPtr^windowVariant=',WindDescPtr^.windowVariant);
	WriteLn ('WindDescPtr^controlCount=',WindDescPtr^.controlCount);
	WriteLn ('WindDescPtr^numItems=',WindDescPtr^.numItems);
	WriteLn ('WindDescPtr^text_name=',WindDescPtr^.text_name);	
END;
{$ENDC} 

{$IFC qDebug}
PROCEDURE TVUAssist.PrintCtrlReq (CtrlReqPtr :SendCtlInfoParamsPtr);
BEGIN
	WriteLn ('CtrlReqPtr^control_rank=', CtrlReqPtr^.control_rank);
	WriteLn ('CtrlReqPtr^window_rank=', CtrlReqPtr^.window_rank);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintCtrlPoint (CtrlReqPtr :SendFindControlParamPtr);
BEGIN
	WriteLn ('CtrlReqPtr=',CtrlReqPtr^.h, CtrlReqPtr^.v);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintCtrlDesc (CtrlDescPtr :MoleCDescPtr);
BEGIN
	Write ('CtrlDescPtr^validFlags=');
	WriteHexInt (CtrlDescPtr^.validFlags);
	WriteLn;
	WriteLn ('CtrlDescPtr^owner_rank=',CtrlDescPtr^.owner_rank);
	WriteLn ('CtrlDescPtr^ctlRect=',CtrlDescPtr^.ctlRect.top,
		CtrlDescPtr^.ctlRect.left, CtrlDescPtr^.ctlRect.bottom,
		CtrlDescPtr^.ctlRect.right);
	WriteLn ('CtrlDescPtr^ctlValue=',CtrlDescPtr^.ctlValue);
	WriteLn ('CtrlDescPtr^ctlMin=',CtrlDescPtr^.ctlMin);
	WriteLn ('CtrlDescPtr^ctlMax=',CtrlDescPtr^.ctlMax);
	WriteLn ('CtrlDescPtr^ctlHilite=',CtrlDescPtr^.ctlHilite);
	WriteLn ('CtrlDescPtr^part=',CtrlDescPtr^.part);
	WriteLn ('CtrlDescPtr^ctlRank=',CtrlDescPtr^.ctlRank);
	WriteLn ('CtrlDescPtr^ctlDefID=',CtrlDescPtr^.ctlDefID);
	WriteLn ('CtrlDescPtr^text_name=',CtrlDescPtr^.text_name);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintDlogReq (DlogReqPtr :SendDItemInfoPtr);
BEGIN
	WriteLn ('DlogReqPtr^item_num=', DlogReqPtr^.item_num);
	WriteLn ('DlogReqPtr^window_rank=', DlogReqPtr^.window_rank);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintDlogDesc (DlogDescPtr :MoleDItemInfoPtr);
BEGIN
	WriteLn ('DlogDescPtr^box=',DlogDescPtr^.box.top,
		DlogDescPtr^.box.left, DlogDescPtr^.box.bottom,
		DlogDescPtr^.box.right);
	WriteLn ('DlogDescPtr^ctlRank=',DlogDescPtr^.ctlRank);
	WriteLn ('DlogDescPtr^itemType=',DlogDescPtr^.itemType);
	WriteLn ('DlogDescPtr^item_enabled=',DlogDescPtr^.item_enabled);
	WriteLn ('DlogDescPtr^has_text=',DlogDescPtr^.has_text);
	WriteLn ('DlogDescPtr^owner_rank=',DlogDescPtr^.owner_rank);
	WriteLn ('DlogDescPtr^item_num=',DlogDescPtr^.item_num);
	WriteLn ('DlogDescPtr^text_name=',DlogDescPtr^.text_name);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintMenuInfoReq (MenuReqPtr :SendMenuInfoParamsPtr);
BEGIN
	WriteLn ('MenuReqPtr^menu_rank=', MenuReqPtr^.menu_rank);
	WriteLn ('MenuReqPtr^menu_ID=', MenuReqPtr^.menu_ID);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintMenuInfoDesc (MenuDescPtr :MoleMenuInfoPtr);
BEGIN
	WriteLn ('MenuDescPtr^num_items=', MenuDescPtr^.num_items);
	WriteLn ('MenuDescPtr^rank=', MenuDescPtr^.rank);
	WriteLn ('MenuDescPtr^left_edge=', MenuDescPtr^.left_edge);
	WriteLn ('MenuDescPtr^menuID=', MenuDescPtr^.menuID);
	WriteLn ('MenuDescPtr^menuWidth=', MenuDescPtr^.menuWidth);
	WriteLn ('MenuDescPtr^menuHeight=', MenuDescPtr^.menuHeight);
	WriteLn ('MenuDescPtr^enableFlags=', MenuDescPtr^.enableFlags);
	WriteLn ('MenuDescPtr^text_name=', MenuDescPtr^.text_name);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintMenuItemReq (MenuReqPtr :SendMenuItemsParamsPtr);
BEGIN
	WriteLn ('MenuReqPtr^menu_rank=', MenuReqPtr^.menu_rank);
	WriteLn ('MenuReqPtr^menu_ID=', MenuReqPtr^.menu_ID);
	WriteLn ('MenuReqPtr^start_item=', MenuReqPtr^.start_item);
	WriteLn ('MenuReqPtr^stop_item=', MenuReqPtr^.stop_item);
END;
{$ENDC}

{$IFC qDebug}
PROCEDURE TVUAssist.PrintMenuItemDesc (MenuDescPtr :MoleDataBlockPtr; OutputSize :integer);
	VAR
		index: 		Size;
BEGIN
	WriteLn ('MenuDescPtr^blockLength=', MenuDescPtr^.blockLength);
	WriteLn ('OutputSize=', OutputSize);
	IF OutputSize > 0 THEN
	BEGIN
		FOR index := 0 TO OutputSize DO
			Write(CHR(integer(MenuDescPtr^.dataBlock[index])));
		WriteLn;
		WriteLn ('MenuDescPtr^dataBlock= ', MenuDescPtr^.dataBlock);
	END
	ELSE
		WriteLn ('MenuDescPtr^dataBlock= **Empty**');
END;
{$ENDC}

{$IFC qDebug}
{$S MAFields}
PROCEDURE TVUAssist.Fields(PROCEDURE DoToField(fieldName: Str255; fieldAddr: Ptr;
	fieldType: INTEGER)); OVERRIDE;
BEGIN
	DoToField('TVUAssist', NIL, bClass);
	DoToField('fMoleRefNum', @fMoleRefNum, bInteger);
	INHERITED Fields(DoToField);
END;
{$ENDC}


{========================================== TGridItem ================================}

{$S VUAssistSeg}
{ TVUAssist.FillDlogItemDesc calls GetExtent on a view to get its rect.  For
grid items, we override it to provide the rect of just the item }
PROCEDURE TGridItem.GetExtent (VAR itsExtent: VRect); OVERRIDE;
VAR
	aCell: GridCell;
BEGIN
	aCell.v:=fRow;
	aCell.h:=fColumn;
	fGridView.CellToVRect (aCell, itsExtent);
END;


{ call the GridView's LocalToWindow }
PROCEDURE TGridItem.LocalToWindow(VAR thePoint: VPoint);
BEGIN
	fGridView.LocalToWindow(thePoint);
END;


{ Free the TGridItem object.  We can't call inherited free, because we never
set up the data structures of our ancestors, because we never called IGridView. }
PROCEDURE TGridItem.Free;
BEGIN
	ShallowFree;
END;


{$IFC qDebug}
PROCEDURE TGridItem.PrintInstanceVars;
BEGIN
	WrLblHexLongint ('TGridItem fGridView=', longint(fGridView));
	WriteLn;
	WriteLn ('TGridItem fColumn=', fColumn);
	WriteLn ('TGridItem fRow=', fRow);
	WriteLn ('TGridItem fRank=', fRank);
	WriteLn ('TGridView fNumOfRows=', fGridView.fNumOfRows);
	WriteLn ('TGridView fNumOfCols=', fGridView.fNumOfCols);
END;
{$ENDC}


{$IFC qDebug}
{$S MAFields}
PROCEDURE TGridItem.Fields(PROCEDURE DoToField(fieldName: Str255; fieldAddr: Ptr;
	fieldType: INTEGER)); OVERRIDE;
BEGIN
	DoToField('TGridItem', NIL, bClass);
	DoToField('fGridView', @fGridView, bObject);
	DoToField('fColumn', @fColumn, bLongint);
	DoToField('fRow', @fRow, bLongint);
	DoToField('fRank', @fRank, bLongint);
	INHERITED Fields(DoToField);
END;
{$ENDC}

