(*+
 *	File:		UVUAssist.p
 *
 *	Contains:	VU assistance class
 *
 *	Written by:	David Shayer
 *
 *	Copyright:	 1988-90 by Apple Computer, Inc., all rights reserved.
 *
 *  Version: 1.0d6
 *
 *	Change History:
 *
 *		  1/10/90	DAS		  new today
 *
 *	To Do:
 *
-*)


{Historical Note:  References to the "Mole" are equivalent to references to "Agent VU".
"Agent VU" was previously called the "Mole".}

(*
So you wanna use VU in your MacApp program?  You need to help the Mole find things
in your windows, since MacApp views appear empty to the Mole.  Fortunately,
the Mole has a hook for an assistance procedure.  This is a routine, in the application
being manipulated by VU, that feeds the Mole inside information about what's in the
views.  This unit implements that assistance routine.

This unit contains 4 files: VUAssist.p, VUAssist.inc1.p, VUAssist.inc2.p,
and VUAssist.a.

You must hook this unit into your MacApp program in the following places:
	In TApplication.IApplication, you must execute this:
		NEW(gVUAssist);
		FailNil(gVUAssist);
		gVUAssist.IVUAssist;
	In TApplication.AboutToLoseControl you must execute this:
		gVUAssist.SuspendMole;
	In TApplication.RegainControl you must execute this:
		gVUAssist.ResumeMole;
When I refer to TApplication, of course I mean your descendent from TApplication.
I think that's all you need to do to hook in the Mole assistance proc.

This is a development release.  If you find any bugs, please report them to me.

You may, if you want, customize this code to support special types of controls
that it doesn't currently support.  However, I reserve the right to change the 
architecture and break your extensions.  Let me state this unambiguously.  I will
be making changes to the architecture, and if you modify the code, I will probably
break you in the next release.  You have been warned.  Support for 
popup menus is coming soon from me.

If you subclass TVUAssist, you will probably create a new global object pointer
to replace gVUAssist.  Be aware that you must keep gVUAssist, and it must 
also point to the object.

Finally, I would like to hear your comments, criticisms, ideas, etc, about my code,
the way I gather info from MacApp, or anything else.
Call me at x44843 or link me at SHAYER1.
-David
*)




UNIT UVUAssist;

INTERFACE

USES

	{  MacApp }
	UMacApp,
{$IFC qDebug}
	UDebug,
{$ENDC}
	
	{ Building Blocks }
	UPrinting, UTEView, UDialog, UGridView,

	{  System Interfaces }
	Errors,
	Resources,
	AppleTalk, ToolUtils,
	OSEvents,
	Packages,
	Script;
	
{$I UVUAssist.inc1.p}

TYPE

	{ There is a duplicate definition of this object in UVUAssist.a.  
	If you update this definition, be sure to update the one in UVUAssist.a. }
	TVUAssist = OBJECT (TObject)
		fMoleRefNum :integer;

		PROCEDURE TVUAssist.IVUAssist;

		PROCEDURE TVUAssist.OpenMoleDriver;

		PROCEDURE TVUAssist.SuspendMole;

		PROCEDURE TVUAssist.ResumeMole; 

		PROCEDURE TVUAssist.SetDebuggerHook (NewHook :ProcPtr); 

		FUNCTION TVUAssist.MoleAssist (Select :integer; Input :Ptr; Output :Ptr; 
			VAR OutputSize :integer; IntResult :integer) :integer; 
		
		FUNCTION TVUAssist.DoMenuInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
			Result :MoleError) :MoleError;

		FUNCTION TVUAssist.DoMenuItems (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
			Result :MoleError) :MoleError;

		FUNCTION TVUAssist.DoFindWindowInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
			Result :MoleError) :MoleError;

		FUNCTION TVUAssist.DoSendWindowInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
			Result :MoleError) :MoleError;

		FUNCTION TVUAssist.DoFindControl (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
			Result :MoleError) :MoleError;

		FUNCTION TVUAssist.DoSendControlInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
			Result :MoleError) :MoleError;

		FUNCTION TVUAssist.DoSendDlogItemInfo (Input :Ptr; Output :Ptr; VAR OutputSize :integer; 
			Result :MoleError) :MoleError;

		PROCEDURE TVUAssist.FillPopupDesc (Popup: TPopup; MenuRank: integer; 
			MenuDescPtr :MoleMenuInfoPtr; VAR OutputSize :integer);

		FUNCTION TVUAssist.FillPopupItemDesc (Popup: TPopup; MenuRank: integer; start: integer; 
			stop: integer; MenuItemDescPtr :MoleDataBlockPtr; VAR OutputSize :integer) :MoleError;

		PROCEDURE TVUAssist.FillCtrlDesc (CtrlDescPtr :MoleCDescPtr; thePart :integer; 
			WindRank :integer; ControlRank :integer; Control :TControl; VAR OutputSize :integer);

		PROCEDURE TVUAssist.ViewRectToWindowRect (VAR aRect :Rect; aView :TView);

		PROCEDURE TVUAssist.FillDlogItemDesc (DlogDescPtr :MoleDItemInfoPtr; 
			Window :TWindow; WindRank :integer; ItemRank :integer; Item :TView; 
			VAR OutputSize :integer);

		FUNCTION TVUAssist.RankToWindow (Rank :integer) :TWindow;
		
		FUNCTION TVUAssist.WindowToRank (WindToFind :WindowPeek) :integer;

		PROCEDURE TVUAssist.GlobalToWindow (Window :TWindow; Pt :Point; VAR VPt :VPoint);

		PROCEDURE TVUAssist.WindowToView (View :TView; VAR VPt :VPoint; VAR Pt :Point);

		FUNCTION TVUAssist.MenuIDToPopup (MenuID :integer) :TPopup;

		FUNCTION TVUAssist.PointToControl (Window :TWindow; Pt :Point; VAR Rank :integer) :TControl;

		FUNCTION TVUAssist.ItemToRank (Window :TWindow; View :TView; 
			DlogViewItems :Boolean) :integer;

		FUNCTION TVUAssist.RankToItem (Window :TWindow; Rank :integer; 
			DlogViewItems :Boolean) :TView;

		FUNCTION TVUAssist.ViewCount (Window :TWindow; DlogViewItems :Boolean) :integer;

		FUNCTION TVUAssist.CountGridItems (theGridView: TGridView): integer;

		FUNCTION TVUAssist.MakeGridItem (theGridView: TGridView; Rank: integer): TGridItem;

		FUNCTION TVUAssist.GetItemType (Item :TView; DlogViewItems :Boolean;
			GridViewItems: Boolean) :ViewItems;

	{$IFC qDebug}
		PROCEDURE TVUAssist.PrintFindWindReq (WindReqPtr :SendFindWindowParamPtr);

		PROCEDURE TVUAssist.PrintWindReq (WindReqPtr :DWindReqPtr);

		PROCEDURE TVUAssist.PrintWindDesc (WindDescPtr :MoleWDescPtr);

		PROCEDURE TVUAssist.PrintCtrlReq (CtrlReqPtr :SendCtlInfoParamsPtr);

		PROCEDURE TVUAssist.PrintCtrlPoint (CtrlReqPtr :SendFindControlParamPtr);

		PROCEDURE TVUAssist.PrintCtrlDesc (CtrlDescPtr :MoleCDescPtr);

		PROCEDURE TVUAssist.PrintDlogReq (DlogReqPtr :SendDItemInfoPtr);

		PROCEDURE TVUAssist.PrintDlogDesc (DlogDescPtr :MoleDItemInfoPtr);

		PROCEDURE TVUAssist.PrintMenuInfoReq (MenuReqPtr :SendMenuInfoParamsPtr);

		PROCEDURE TVUAssist.PrintMenuInfoDesc (MenuDescPtr :MoleMenuInfoPtr);

		PROCEDURE TVUAssist.PrintMenuItemReq (MenuReqPtr :SendMenuItemsParamsPtr);

		PROCEDURE TVUAssist.PrintMenuItemDesc (MenuDescPtr :MoleDataBlockPtr; OutputSize :integer);

		PROCEDURE TVUAssist.Fields(PROCEDURE DoToField(fieldName: Str255; fieldAddr: Ptr;
			fieldType: INTEGER)); OVERRIDE;
	{$ENDC}
	END;

	TGridItem = OBJECT (TGridView)
		fGridView:	TGridView;
		fColumn:	longint;		{ 1 based counter }
		fRow:		longint;		{ 1 based counter }
		fRank:		longint;		{ (row-1) * rowlength + column }

	PROCEDURE TGridItem.GetExtent(VAR itsExtent: VRect); OVERRIDE;

	PROCEDURE TGridItem.LocalToWindow(VAR thePoint: VPoint); OVERRIDE;
	
	PROCEDURE TGridItem.Free; OVERRIDE;

	{$IFC qDebug}
		PROCEDURE TGridItem.PrintInstanceVars;

		PROCEDURE TGridItem.Fields(PROCEDURE DoToField(fieldName: Str255; fieldAddr: Ptr;
			fieldType: INTEGER)); OVERRIDE;
	{$ENDC}
	END;

VAR
	gVUAssist :					TVUAssist;		{ the mole assist proc }
	{$IFC qDebug}
		gVUDebug :				Boolean;		{ debug mode is on }
	{$ENDC}


IMPLEMENTATION

{$I UVUAssist.inc2.p}

END.
